package javax.security.auth;

import java.text.MessageFormat;
import java.util.Objects;
import java.util.Collection;
import java.util.ListIterator;
import java.util.LinkedList;
import java.io.ObjectStreamField;
import java.util.AbstractSet;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Iterator;
import java.util.HashSet;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.security.AccessController;
import java.security.DomainCombiner;
import java.security.PrivilegedAction;
import java.security.AccessControlContext;
import sun.security.util.ResourcesMgr;
import java.util.Collections;
import java.security.ProtectionDomain;
import java.security.Principal;
import java.util.Set;
import java.io.Serializable;

public final class Subject implements Serializable
{
    private static final long serialVersionUID = -8308522755600156056L;
    Set<Principal> principals;
    transient Set<Object> pubCredentials;
    transient Set<Object> privCredentials;
    private volatile boolean readOnly;
    private static final int PRINCIPAL_SET = 1;
    private static final int PUB_CREDENTIAL_SET = 2;
    private static final int PRIV_CREDENTIAL_SET = 3;
    private static final ProtectionDomain[] NULL_PD_ARRAY;
    
    public Subject() {
        this.readOnly = false;
        this.principals = Collections.synchronizedSet(new Subject.SecureSet(this, 1));
        this.pubCredentials = Collections.synchronizedSet(new Subject.SecureSet(this, 2));
        this.privCredentials = Collections.synchronizedSet(new Subject.SecureSet(this, 3));
    }
    
    public Subject(boolean readOnly, Set<? extends Principal> set, Set<?> set2, Set<?> set3) {
        this.readOnly = false;
        if (set == null || set2 == null || set3 == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.input.s."));
        }
        this.principals = Collections.synchronizedSet(new Subject.SecureSet(this, 1, set));
        this.pubCredentials = Collections.synchronizedSet(new Subject.SecureSet(this, 2, set2));
        this.privCredentials = Collections.synchronizedSet(new Subject.SecureSet(this, 3, set3));
        this.readOnly = readOnly;
        return;
    }
    
    public void setReadOnly() {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Subject.AuthPermissionHolder.SET_READ_ONLY_PERMISSION);
        }
        this.readOnly = true;
    }
    
    public boolean isReadOnly() {
        return this.readOnly;
    }
    
    public static Subject getSubject(final AccessControlContext accessControlContext) {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Subject.AuthPermissionHolder.GET_SUBJECT_PERMISSION);
        }
        if (accessControlContext == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.AccessControlContext.provided"));
        }
        return (Subject)AccessController.doPrivileged(new PrivilegedAction<Subject>() {
            @Override
            public Subject run() {
                DomainCombiner domainCombiner = accessControlContext.getDomainCombiner();
                if (!(domainCombiner instanceof SubjectDomainCombiner)) {
                    return null;
                }
                return ((SubjectDomainCombiner)domainCombiner).getSubject();
            }
        });
    }
    
    public static <T> T doAs(Subject subject, PrivilegedAction<T> action) {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Subject.AuthPermissionHolder.DO_AS_PERMISSION);
        }
        if (action == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.action.provided"));
        }
        return AccessController.doPrivileged(action, createContext(subject, AccessController.getContext()));
    }
    
    public static <T> T doAs(Subject subject, PrivilegedExceptionAction<T> action) throws PrivilegedActionException {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Subject.AuthPermissionHolder.DO_AS_PERMISSION);
        }
        if (action == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.action.provided"));
        }
        return AccessController.doPrivileged(action, createContext(subject, AccessController.getContext()));
    }
    
    public static <T> T doAsPrivileged(Subject subject, PrivilegedAction<T> action, AccessControlContext accessControlContext) {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Subject.AuthPermissionHolder.DO_AS_PRIVILEGED_PERMISSION);
        }
        if (action == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.action.provided"));
        }
        accessControlContext = ((accessControlContext == null) ? new AccessControlContext(Subject.NULL_PD_ARRAY) : accessControlContext);
        return AccessController.doPrivileged(action, createContext(subject, accessControlContext));
    }
    
    public static <T> T doAsPrivileged(Subject subject, PrivilegedExceptionAction<T> action, AccessControlContext accessControlContext) throws PrivilegedActionException {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Subject.AuthPermissionHolder.DO_AS_PRIVILEGED_PERMISSION);
        }
        if (action == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.action.provided"));
        }
        accessControlContext = ((accessControlContext == null) ? new AccessControlContext(Subject.NULL_PD_ARRAY) : accessControlContext);
        return AccessController.doPrivileged(action, createContext(subject, accessControlContext));
    }
    
    private static AccessControlContext createContext(final Subject subject, final AccessControlContext accessControlContext) {
        return (AccessControlContext)AccessController.doPrivileged(new PrivilegedAction<AccessControlContext>() {
            @Override
            public AccessControlContext run() {
                if (subject == null) {
                    return new AccessControlContext(accessControlContext, null);
                }
                return new AccessControlContext(accessControlContext, new SubjectDomainCombiner(subject));
            }
        });
    }
    
    public Set<Principal> getPrincipals() {
        return this.principals;
    }
    
    public <T extends Principal> Set<T> getPrincipals(Class<T> clazz) {
        if (clazz == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.Class.provided"));
        }
        return new Subject.ClassSet(1, clazz);
    }
    
    public Set<Object> getPublicCredentials() {
        return this.pubCredentials;
    }
    
    public Set<Object> getPrivateCredentials() {
        return this.privCredentials;
    }
    
    public <T> Set<T> getPublicCredentials(Class<T> clazz) {
        if (clazz == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.Class.provided"));
        }
        return new Subject.ClassSet(2, clazz);
    }
    
    public <T> Set<T> getPrivateCredentials(Class<T> clazz) {
        if (clazz == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.Class.provided"));
        }
        return new Subject.ClassSet(3, clazz);
    }
    
    @Override
    public boolean equals(Object o) {
        if (o == null) {
            return false;
        }
        if (this == o) {
            return true;
        }
        if ((o instanceof Subject)) {
            Subject subject = (Subject)o;
            HashSet<Object> set;
            synchronized (subject.principals) {
                set = new HashSet(subject.principals);
            }
            if (!this.principals.equals(set)) {
                return false;
            }
            Set<Object> set2;
            synchronized (subject.pubCredentials) {
                set2 = new HashSet(subject.pubCredentials);
            }
            if (!this.pubCredentials.equals(set2)) {
                return false;
            }
            Set set3;
            synchronized (subject.privCredentials) {
                set3 = new HashSet(subject.privCredentials);
            }
            return this.privCredentials.equals(set3);
        }
        else {
            return false;
        }
    }
    
    @Override
    public String toString() {
        return this.toString(true);
    }
    
    String toString(boolean b) {
        String string = ResourcesMgr.getString("Subject.");
        String str = "";
        synchronized (this.principals) {
            Iterator<Principal> iterator = this.principals.iterator();
            while (iterator.hasNext()) {
                str = str + ResourcesMgr.getString(".Principal.") + ((Principal)iterator.next()).toString() + ResourcesMgr.getString("NEWLINE");
            }
        }
        synchronized (this.pubCredentials) {
            Iterator<Object> iterator2 = this.pubCredentials.iterator();
            while (iterator2.hasNext()) {
                str = str + ResourcesMgr.getString(".Public.Credential.") + iterator2.next().toString() + ResourcesMgr.getString("NEWLINE");
            }
        }
        if (b) {
            synchronized (this.privCredentials) {
                Iterator<Object> iterator3 = this.privCredentials.iterator();
                while (iterator3.hasNext()) {
                    try {
                        str = str + ResourcesMgr.getString(".Private.Credential.") + iterator3.next().toString() + ResourcesMgr.getString("NEWLINE");
                    }
                    catch (SecurityException ex) {
                        str += ResourcesMgr.getString(".Private.Credential.inaccessible.");
                        break;
                    }
                }
            }
        }
        return string + str;
    }
    
    @Override
    public int hashCode() {
        int n = 0;
        synchronized (this.principals) {
            Iterator<Principal> iterator = this.principals.iterator();
            while (iterator.hasNext()) {
                n ^= ((Principal)iterator.next()).hashCode();
            }
        }
        synchronized (this.pubCredentials) {
            Iterator<Object> iterator2 = this.pubCredentials.iterator();
            while (iterator2.hasNext()) {
                n ^= this.getCredHashCode(iterator2.next());
            }
        }
        return n;
    }
    
    private int getCredHashCode(Object o) {
        try {
            return o.hashCode();
        }
        catch (IllegalStateException ex) {
            return o.getClass().toString().hashCode();
        }
    }
    
    private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
        synchronized (this.principals) {
            objectOutputStream.defaultWriteObject();
        }
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
        ObjectInputStream.GetField fields = objectInputStream.readFields();
        this.readOnly = fields.get("readOnly", false);
        Set<Object> set = (Set)fields.get("principals", null);
        if (set == null) {
            throw new NullPointerException(ResourcesMgr.getString("invalid.null.input.s."));
        }
        try {
            this.principals = Collections.synchronizedSet(new Subject.SecureSet(this, 1, set));
        }
        catch (NullPointerException ex) {
            this.principals = Collections.synchronizedSet(new Subject.SecureSet(this, 1));
        }
        this.pubCredentials = Collections.synchronizedSet(new Subject.SecureSet(this, 2));
        this.privCredentials = Collections.synchronizedSet(new Subject.SecureSet(this, 3));
        return;
    }
    
    static {
        NULL_PD_ARRAY = new ProtectionDomain[0];
    }
    
    private static class SecureSet<E> extends AbstractSet<E> implements Serializable
    {
        private static final long serialVersionUID = 7911754171111800359L;
        private static final ObjectStreamField[] serialPersistentFields;
        Subject subject;
        LinkedList<E> elements;
        private int which;
        
        SecureSet(Subject subject, int which) {
            this.subject = subject;
            this.which = which;
            this.elements = new LinkedList();
        }
        
        SecureSet(Subject subject, int which, Set<? extends E> c) {
            this.subject = subject;
            this.which = which;
            this.elements = new LinkedList(c);
        }
        
        @Override
        public int size() {
            return this.elements.size();
        }
        
        @Override
        public Iterator<E> iterator() {
            return new Iterator<E>() {
                ListIterator<E> i = list.listIterator(0);
                
                @Override
                public boolean hasNext() {
                    return this.i.hasNext();
                }
                
                @Override
                public E next() {
                    if (Subject.SecureSet.this.which != 3) {
                        return this.i.next();
                    }
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        try {
                            securityManager.checkPermission(new PrivateCredentialPermission(list.get(this.i.nextIndex()).getClass().getName(), Subject.SecureSet.this.subject.getPrincipals()));
                        }
                        catch (SecurityException ex) {
                            this.i.next();
                            throw ex;
                        }
                    }
                    return this.i.next();
                }
                
                @Override
                public void remove() {
                    if (Subject.SecureSet.this.subject.isReadOnly()) {
                        throw new IllegalStateException(ResourcesMgr.getString("Subject.is.read.only"));
                    }
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        switch (Subject.SecureSet.this.which) {
                            case 1: {
                                securityManager.checkPermission(Subject.AuthPermissionHolder.MODIFY_PRINCIPALS_PERMISSION);
                                break;
                            }
                            case 2: {
                                securityManager.checkPermission(Subject.AuthPermissionHolder.MODIFY_PUBLIC_CREDENTIALS_PERMISSION);
                                break;
                            }
                            default: {
                                securityManager.checkPermission(Subject.AuthPermissionHolder.MODIFY_PRIVATE_CREDENTIALS_PERMISSION);
                                break;
                            }
                        }
                    }
                    this.i.remove();
                    return;
                }
            };
        }
        
        @Override
        public boolean add(E e) {
            if (this.subject.isReadOnly()) {
                throw new IllegalStateException(ResourcesMgr.getString("Subject.is.read.only"));
            }
            SecurityManager securityManager = System.getSecurityManager();
            if (securityManager != null) {
                switch (this.which) {
                    case 1: {
                        securityManager.checkPermission(Subject.AuthPermissionHolder.MODIFY_PRINCIPALS_PERMISSION);
                        break;
                    }
                    case 2: {
                        securityManager.checkPermission(Subject.AuthPermissionHolder.MODIFY_PUBLIC_CREDENTIALS_PERMISSION);
                        break;
                    }
                    default: {
                        securityManager.checkPermission(Subject.AuthPermissionHolder.MODIFY_PRIVATE_CREDENTIALS_PERMISSION);
                        break;
                    }
                }
            }
            switch (this.which) {
                case 1: {
                    if (!(e instanceof Principal)) {
                        throw new SecurityException(ResourcesMgr.getString("attempting.to.add.an.object.which.is.not.an.instance.of.java.security.Principal.to.a.Subject.s.Principal.Set"));
                    }
                    break;
                }
            }
            return !this.elements.contains(e) && this.elements.add(e);
        }
        
        @Override
        public boolean remove(Object obj) {
            Iterator<E> iterator = this.iterator();
            while (iterator.hasNext()) {
                Object o;
                if (this.which != 3) {
                    o = iterator.next();
                }
                else {
                    o = AccessController.doPrivileged(new PrivilegedAction<E>() {
                        @Override
                        public E run() {
                            return iterator.next();
                        }
                    });
                }
                if (o == null) {
                    if (obj == null) {
                        iterator.remove();
                        return true;
                    }
                    continue;
                }
                else {
                    if (o.equals(obj)) {
                        iterator.remove();
                        return true;
                    }
                    continue;
                }
            }
            return false;
        }
        
        @Override
        public boolean contains(Object obj) {
            Iterator<E> iterator = this.iterator();
            while (iterator.hasNext()) {
                Object o;
                if (this.which != 3) {
                    o = iterator.next();
                }
                else {
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        securityManager.checkPermission(new PrivateCredentialPermission(obj.getClass().getName(), this.subject.getPrincipals()));
                    }
                    o = AccessController.doPrivileged(new PrivilegedAction<E>() {
                        @Override
                        public E run() {
                            return iterator.next();
                        }
                    });
                }
                if (o == null) {
                    if (obj == null) {
                        return true;
                    }
                    continue;
                }
                else {
                    if (o.equals(obj)) {
                        return true;
                    }
                    continue;
                }
            }
            return false;
        }
        
        @Override
        public boolean removeAll(Collection<?> obj) {
            Objects.requireNonNull(obj);
            boolean n = false;
            for (Iterator iterator = this.iterator(); iterator.hasNext(); n = true) {
                Object o;
                if (this.which != 3) {
                    o = iterator.next();
                }
                else {
                    o = AccessController.doPrivileged(new PrivilegedAction<E>() {
                        @Override
                        public E run() {
                            return iterator.next();
                        }
                    });
                }
                for (Object next : obj) {
                    if (o == null) {
                        if (next == null) {
                            iterator.remove();
                            break;
                        }
                        continue;
                    }
                    else {
                        if (o.equals(next)) {
                            iterator.remove();
                            break;
                        }
                        continue;
                    }
                }
            }
            return n;
        }
        
        @Override
        public boolean retainAll(Collection<?> obj) {
            Objects.requireNonNull(obj);
            boolean n = false;
            Iterator<E> iterator = this.iterator();
            while (iterator.hasNext()) {
                boolean n2 = false;
                Object o;
                if (this.which != 3) {
                    o = iterator.next();
                }
                else {
                    o = AccessController.doPrivileged(new PrivilegedAction<E>() {
                        @Override
                        public E run() {
                            return iterator.next();
                        }
                    });
                }
                for (Object next : obj) {
                    if (o == null) {
                        if (next == null) {
                            n2 = true;
                            break;
                        }
                        continue;
                    }
                    else {
                        if (o.equals(next)) {
                            n2 = true;
                            break;
                        }
                        continue;
                    }
                }
                if (n2 == 0) {
                    iterator.remove();
                    n = true;
                }
                else {
                    continue;
                }
            }
            return n;
        }
        
        @Override
        public void clear() {
            Iterator<E> iterator = this.iterator();
            while (iterator.hasNext()) {
                if (this.which != 3) {
                    iterator.next();
                }
                else {
                    AccessController.doPrivileged(new PrivilegedAction<E>() {
                        @Override
                        public E run() {
                            return iterator.next();
                        }
                    });
                }
                iterator.remove();
            }
        }
        
        private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
            if (this.which == 3) {
                Iterator<E> iterator = this.iterator();
                while (iterator.hasNext()) {
                    iterator.next();
                }
            }
            ObjectOutputStream.PutField putFields = objectOutputStream.putFields();
            putFields.put("this$0", this.subject);
            putFields.put("elements", this.elements);
            putFields.put("which", this.which);
            objectOutputStream.writeFields();
        }
        
        private void readObject(ObjectInputStream objectInputStream) throws IOException, ClassNotFoundException {
            ObjectInputStream.GetField fields = objectInputStream.readFields();
            this.subject = (Subject)fields.get("this$0", null);
            this.which = fields.get("which", 0);
            LinkedList<Object> list = (LinkedList)fields.get("elements", null);
            if (list.getClass() != LinkedList.class) {
                this.elements = new LinkedList(list);
            }
            else {
                this.elements = list;
            }
        }
        
        static {
            serialPersistentFields = new ObjectStreamField[] { new ObjectStreamField("this$0", Subject.class), new ObjectStreamField("elements", LinkedList.class), new ObjectStreamField("which", Integer.TYPE) };
        }
    }
    
    private class ClassSet<T> extends AbstractSet<T>
    {
        private int which;
        private Class<T> c;
        private Set<T> set;
        
        ClassSet(int which, Class<T> c) {
            this.which = which;
            this.c = c;
            this.set = new HashSet();
            switch (which) {
                case 1: {
                    synchronized (Subject.this.principals) {
                        this.populateSet();
                    }
                    break;
                }
                case 2: {
                    synchronized (Subject.this.pubCredentials) {
                        this.populateSet();
                    }
                    break;
                }
                default: {
                    synchronized (Subject.this.privCredentials) {
                        this.populateSet();
                        break;
                    }
                    break;
                }
            }
        }
        
        private void populateSet() {
            Iterator<Object> iterator;
            switch (this.which) {
                case 1: {
                    iterator = Subject.this.principals.iterator();
                    break;
                }
                case 2: {
                    iterator = Subject.this.pubCredentials.iterator();
                    break;
                }
                default: {
                    iterator = Subject.this.privCredentials.iterator();
                    break;
                }
            }
            while (iterator.hasNext()) {
                Object o;
                if (this.which == 3) {
                    o = AccessController.doPrivileged(new PrivilegedAction<Object>() {
                        @Override
                        public Object run() {
                            return iterator.next();
                        }
                    });
                }
                else {
                    o = iterator.next();
                }
                if (this.c.isAssignableFrom(o.getClass())) {
                    if (this.which != 3) {
                        this.set.add(o);
                    }
                    else {
                        SecurityManager securityManager = System.getSecurityManager();
                        if (securityManager != null) {
                            securityManager.checkPermission(new PrivateCredentialPermission(o.getClass().getName(), Subject.this.getPrincipals()));
                        }
                        this.set.add(o);
                    }
                }
                else {
                    continue;
                }
            }
        }
        
        @Override
        public int size() {
            return this.set.size();
        }
        
        @Override
        public Iterator<T> iterator() {
            return this.set.iterator();
        }
        
        @Override
        public boolean add(T t) {
            if (!t.getClass().isAssignableFrom(this.c)) {
                throw new SecurityException(new MessageFormat(ResourcesMgr.getString("attempting.to.add.an.object.which.is.not.an.instance.of.class")).format(new Object[] { this.c.toString() }));
            }
            return this.set.add(t);
        }
    }
    
    static class AuthPermissionHolder
    {
        static final AuthPermission DO_AS_PERMISSION;
        static final AuthPermission DO_AS_PRIVILEGED_PERMISSION;
        static final AuthPermission SET_READ_ONLY_PERMISSION;
        static final AuthPermission GET_SUBJECT_PERMISSION;
        static final AuthPermission MODIFY_PRINCIPALS_PERMISSION;
        static final AuthPermission MODIFY_PUBLIC_CREDENTIALS_PERMISSION;
        static final AuthPermission MODIFY_PRIVATE_CREDENTIALS_PERMISSION;
        
        static {
            DO_AS_PERMISSION = new AuthPermission("doAs");
            DO_AS_PRIVILEGED_PERMISSION = new AuthPermission("doAsPrivileged");
            SET_READ_ONLY_PERMISSION = new AuthPermission("setReadOnly");
            GET_SUBJECT_PERMISSION = new AuthPermission("getSubject");
            MODIFY_PRINCIPALS_PERMISSION = new AuthPermission("modifyPrincipals");
            MODIFY_PUBLIC_CREDENTIALS_PERMISSION = new AuthPermission("modifyPublicCredentials");
            MODIFY_PRIVATE_CREDENTIALS_PERMISSION = new AuthPermission("modifyPrivateCredentials");
        }
    }
}
